
/*
 * device user library
 */

#ifndef _GISI_GISU_D_HXX_
#define _GISI_GISU_D_HXX_

#include "gpibc/gumi.hxx"
#include "cplRaw/sm.hxx"
#include "umi/umi_user.hxx"
#include <iostream>
#include <sstream>

/*
 class GpibDevice implements a IEEE488.2 compatible IO
 solution for GPIB Talker/Listener devices (instruments)
 on top of the INES GPIB device driver.

 The implementor must provide a event-driven 
 MyInstrument::Intrument class. 
 
 */

namespace INdepGpib100 {
  namespace Gif=INdepGpib100;

  
  /* 
   * messages send to the instrument parser
   */
  enum {
    M_DAB=255,		/* reserve 0-255 for DAB	*/
    M_GET,		/* GET received			*/		
    M_INITIALIZE,	/* Device clear			*/
    M_UNTERMINATED,	/* 				*/
    M_INTERRUPTED,	/* 				*/
    M_END,		/* End of program message	*/
    M_STATUS,		/* status update		*/
    M_FAIL,		/* IO failure			*/
    M_LAST
  };

  class GpibDevice;
  class Instrument {
   public:

    /* 
     * must return after command execution 
     * or initialization
     */
    virtual int parse(GpibDevice *pgd)=0;

    Instrument() {}
    virtual ~Instrument() {}
  };

  class GpibDevice:public std::streambuf, public Failable, private SmM {
    Umi::Umi *_pum;
    int _eErr;
    Umi::Umi *pum() { return _pum; }

    enum {
      S_CLOSED,
      S_IDLE,
      S_INIT,
      S_READ,
      S_SENT,
      CNT_STATES
    };

    Failable &setResponseState( GpibResp *pgh, const char *sz, int i ) { 
      if ((_eErr=(ErrE)pgh->eErr())) return setFailed(sz, i);
      else return *this; 
    }


    /* input buffer */
    char *_pchBuf;
    char *_pchBufLim;
    int  _fLastIn;
    int  _fLastOut;
    int  _eEof;
    GpibDeviceStatus _gds;

    /* output buffer */

    /*
     * input request
     */
    virtual int underflow();				
    virtual int overflow(int ch);				
    virtual int sync();				

    int setEof(int e) { _eEof= e; return EOF; }
    bool isEof() { return _eEof != 0; }

    void set( const GpibDeviceStatus &gds ) { _gds=gds; }
  
   public:
    int getEof() { int e(_eEof); _eEof=0; return e; }

    /* write output queue END mode - reset by sync() */
    void setEndNL() { _fLastOut=10; }	/* END on NL - ONLY ONE NL immediately sync'ed  */
    void setEndNext() { _fLastOut=1; }	/* END at this block	*/
    void clearEnd() { _fLastOut=0; }	

    /*
     * interface status flags
     */
    Gif::StatE fStat() { return Gif::StatE(_gds.fStat()); }
    bool isListen() { return _gds.fStat() & F_LA; }	/* LADS || LACS			*/
    bool isTalk() { return _gds.fStat() & F_TA; }	/* TADS || TACS			*/


    /*
     * Status Byte and Service Request Generation
     * [IEEE488.2 Sect 11]
     *
     * The MAV (bit 4) is or'ed in to STB by the driver if
     * there is data pending to be send.
     *
     * If you don't want bit 4 or'ed in this way, 
     * set bit 4 in fForce and clear it in fStb.
     * 
     * Set in fSre all bits which shall request service,
     * i.e. set SRQ active.
     * 
     * At the time setStatus returns, the STB and SRQ 
     * need not necessarily be updated at the 
     * interface hardware.
     */
    Failable& setStatus( int fStb, int fSre, int fForce=0 );

    /*
     * run the instrument, returns by failure or "power off"
     */
    Failable &run(Instrument *);

    /*
     * pon 
     */
    Failable &open( bool fSc, int iPa, bool fSa, int iSa );

    /*
     * close - synchronously drain the input queue
     */
    Failable &close( );

    GpibDevice( Failable *pfa, const Umi::Factory *puf);
    ~GpibDevice() ;
  };

}





#if (0)
#include "gpibc/gumi.hxx"
#include "cplRaw/sm.hxx"
#include "umi/umi_user.hxx"
#include <iostream>
#include <sstream>

namespace INdepGpib100 {

  /*
   * GPIB user base class ( device mode )
   *
   * provides a framework for the implementation 
   * of IEEE488.2 compliant devices
   *
   */
  class GpibDevice:public Umi::UmiUser, private SmM {
    //int _iTrans;

    enum {
      S_IDLE,
      S_INIT
    };

    int _eErr;
    GpibDeviceStatus _gds;

    char *_pchBuf;
    char *_pch;
    char *_pchLim;
    char *_pchLimBuf;


    void initBuf(const char *, size_t);




    //void setTransNext() { _iTrans ++ ; }
    //int iTransCur() { return _iTrans; }			/* current transaction id	*/

    const GpibDevice &setFailed( Umi::Umi *pum ) {  return *this; }	/* UMI failed		*/
    const GpibDevice &setFailed( ErrE eErr ) { _eErr=eErr; return *this; }

    //std::streambuf *psbIn() { return  &_sb; }

    void set( const GpibDeviceStatus &gds ) { _gds=gds; }
    GpibDevice &setState( GpibResp *pgh ) { _eErr=(ErrE)pgh->eErr(); return *this; }

    /*
     * message wait
     */
    GpibDevice &getMsg( Umi::UmiControlBuf* pUcb, Umi::UmiDataBuf *pUdb);

    /*
     * message processor, must be called after getMsg() return
     */
    GpibDevice &processMsg( Umi::UmiControlBuf* pUcb, Umi::UmiDataBuf *pUdb);

    class InputBuffer {
      bool _fCmd;		/* command/data	*/
      char *_pchBuf;
      char *_pch;
      char *_pchLim;
      char *_pchLimBuf;
    };

   public:
    bool isGood() const { return _eErr == E_OK && !isBad(); }	/* fine 		*/
    bool isFail() const { return _eErr != E_OK || isBad(); }	/* bad request	 	*/
    int eErr() const { return _eErr; }

    bool operator!() const { return isFail(); }

    GpibDevice &run( );


    /*
     * input buffer access
     */
    int underflow(bool fBump);				/* fill the buffer		*/

    int sgetc() {
      if (_pch >= _pchLim) return underflow(false);
      else return *(unsigned char *)_pch;
    }

    int sbumpc() {
      if (_pch >= _pchLim) return underflow(true);
      else return *(unsigned char *)(_pch++);
    }

    StatE fStat() { return StatE(_gds.fStat()); }
    bool isListen() { return _gds.fStat() & F_LA; }	/* LADS || LACS			*/
    bool isTalk() { return _gds.fStat() & F_TA; }	/* TADS || TACS			*/

    

    /*
     * Status Byte and Service Request Generation
     * [IEEE488.2 Sect 11]
     *
     * The MAV (bit 4) is or'ed in to STB by the driver if
     * there is data pending to be send at the drivers end. 
     *
     * If you don't want bit 4 or'ed in that way, 
     * set bit 4 in fForce and clear it in fStb.
     * 
     * Set in fSre all bits which shall request service,
     * i.e. set SRQ active.
     * 
     * At the time setStatus returns, the STB and SRQ 
     * need not necessarily be updated at the 
     * interface hardware.
     */
    void setStatus( int fStb, int fSre, int fForce=0 );



    /*
     * query results, sets the <query> message true and
     * sets output queue data
     */
    void query( const char *mpch, size_t s, bool fEnd=true);

    /*
     * parser functions
     */
    //virtual void parserInput( const char *mpch, size_t s );
    //virtual void parserTrigger( );		
    //virtual void parserReset( );

    GpibDevice &open( bool fSc, int iPa, bool fSa, int iSa );

    /*
     * close - synchronously drain the input queue
     */
    GpibDevice &close( );

    GpibDevice( Failable *pfa, const Umi::Factory *puf, int fDebug=0);
    ~GpibDevice() ;
  };

}
#endif
#endif
